@@ -1,9 +1,10 @@
 import enum
 import mmap
-import functools
+import sys
 
 from xkbcommon._ffi import ffi, lib
 
+
 class _keepref:
     """Function wrapper that keeps a reference to another object."""
     def __init__(self, ref, func):
@@ -13,14 +14,31 @@ class _keepref:
     def __call__(self, *args, **kwargs):
         self.func(*args, **kwargs)
 
+
+# enum.IntFlag was changed in a non-backward-compatible manner in
+# Python 3.11: the __str__ method was changed to int.__str__(). We
+# retain the old behaviour because it is useful, and in order not to
+# surprise existing users of python-xkbcommon
+
+if sys.version_info < (3, 11):
+    _IntFlag = enum.IntFlag
+else:
+    class _IntFlag(int, enum.Flag, boundary=enum.KEEP):
+        """IntFlag implementation that retains the pre-python-3.11 behaviour
+        """
+        pass
+
+
 class XKBError(Exception):
     """Base for all XKB exceptions"""
     pass
 
+
 class XKBPathError(Exception):
     """There was a problem altering the include path of a Context."""
     pass
 
+
 class XKBBufferTooSmall(XKBError):
     """A buffer created for libxkbcommon to return data in was too small.
 
@@ -29,20 +47,37 @@ class XKBBufferTooSmall(XKBError):
     """
     pass
 
+
 class XKBInvalidKeysym(XKBError):
     pass
 
+
 class XKBKeymapCreationFailure(XKBError):
     """Unable to create a keymap."""
     pass
 
+
 class XKBKeymapReadError(XKBError):
     """Unable to fetch a keymap as a string."""
     pass
 
+
 class XKBInvalidModifierIndex(XKBError):
     pass
 
+
+class XKBInvalidKeycode(XKBError):
+    pass
+
+
+class XKBKeyDoesNotExist(XKBError):
+    """A given key name does not exist.
+    """
+    def __init__(self, key_name):
+        super().__init__(key_name)
+        self.key_name = key_name
+
+
 class XKBModifierDoesNotExist(XKBError):
     """A given modifier name does not exist.
 
@@ -54,36 +89,45 @@ class XKBModifierDoesNotExist(XKBError):
         super().__init__(modifier_name)
         self.modifier_name = modifier_name
 
+
 class XKBInvalidLayoutIndex(XKBError):
     pass
 
+
 class XKBLayoutDoesNotExist(XKBError):
     def __init__(self, index_name):
         super().__init__(index_name)
         self.index_name = index_name
 
+
 class XKBInvalidLEDIndex(XKBError):
     pass
 
+
 class XKBLEDDoesNotExist(XKBError):
     def __init__(self, led_name):
         super().__init__(led_name)
         self.led_name = led_name
 
+
 # Internal helper for logging callback
 def _onerror_do_nothing(exception, exc_value, traceback):
     return
+
+
 @ffi.def_extern(onerror=_onerror_do_nothing)
 def _log_handler(user_data, level, message):
     context = ffi.from_handle(user_data)
     if context._log_fn:
         context._log_fn(context, level, ffi.string(message).decode('utf8'))
 
+
 # Internal helper for keycode iteration
 @ffi.def_extern(onerror=_onerror_do_nothing)
 def _key_for_each_helper(keymap, key, data):
-    l = ffi.from_handle(data)
-    l.append(key)
+    k = ffi.from_handle(data)
+    k.append(key)
+
 
 # Keysyms http://xkbcommon.org/doc/current/group__keysyms.html
 
@@ -97,6 +141,7 @@ def keysym_get_name(keysym):
         raise XKBBufferTooSmall()
     return ffi.string(name).decode('ascii')
 
+
 def keysym_from_name(name, case_insensitive=False):
     "Get a keysym from its name."
     flags = 0
@@ -104,6 +149,7 @@ def keysym_from_name(name, case_insensitive=False):
         flags = flags | lib.XKB_KEYSYM_CASE_INSENSITIVE
     return lib.xkb_keysym_from_name(name.encode('ascii'), flags)
 
+
 def keysym_to_string(keysym):
     buffer = ffi.new("char[64]")
     r = lib.xkb_keysym_to_utf8(keysym, buffer, len(buffer))
@@ -113,6 +159,27 @@ def keysym_to_string(keysym):
         return
     return ffi.string(buffer).decode('utf8')
 
+
+def keysym_to_upper(keysym):
+    """Convert a keysym to its uppercase form.
+
+    If there is no such form, the keysym is returned unchanged.
+
+    The conversion rules may be incomplete; prefer to work with the
+    Unicode representation instead, when possible.
+    """
+    return lib.xkb_keysym_to_upper(keysym)
+
+
+def keysym_to_lower(keysym):
+    """Convert a keysym to its lowercase form.
+
+    The conversion rules may be incomplete; prefer to work with the
+    Unicode representation instead, when possible.
+    """
+    return lib.xkb_keysym_to_lower(keysym)
+
+
 # Library Context http://xkbcommon.org/doc/current/group__context.html
 
 class Context:
@@ -198,7 +265,7 @@ class Context:
     def num_include_paths(self):
         """Return the number of include paths"""
         return lib.xkb_context_num_include_paths(self._context)
-        
+
     def include_path(self):
         "Iterate over the include path."
         i = 0
@@ -320,7 +387,7 @@ class Context:
         "Create a Keymap from an open file"
         try:
             fn = file.fileno()
-        except:
+        except Exception:
             load_method = "read_file"
             keymap = file.read()
             r = lib.xkb_keymap_new_from_string(
@@ -346,7 +413,8 @@ class Context:
             self, string, format=lib.XKB_KEYMAP_FORMAT_TEXT_V1):
         "Create a Keymap from a keymap string."
         r = lib.xkb_keymap_new_from_string(
-            self._context, string.encode("ascii"), format, lib.XKB_KEYMAP_COMPILE_NO_FLAGS)
+            self._context, string.encode("ascii"), format,
+            lib.XKB_KEYMAP_COMPILE_NO_FLAGS)
         if r == ffi.NULL:
             raise XKBKeymapCreationFailure(
                 "xkb_keymap_new_from_string returned NULL")
@@ -364,6 +432,7 @@ class Context:
                 "xkb_keymap_new_from_buffer returned NULL")
         return Keymap(self, r, "buffer")
 
+
 class Keymap:
     """A keymap.
 
@@ -391,7 +460,7 @@ class Keymap:
         kms = ffi.string(r)
         lib.free(r)
         return kms
-        
+
     def get_as_string(self, format=lib.XKB_KEYMAP_FORMAT_TEXT_V1):
         """Get the compiled keymap as a string.
 
@@ -431,6 +500,18 @@ class Keymap:
             self._get_valid_keycodes()
         return iter(self._valid_keycodes)
 
+    def key_get_name(self, key):
+        r = lib.xkb_keymap_key_get_name(self._keymap, key)
+        if r == ffi.NULL:
+            raise XKBInvalidKeycode
+        return ffi.string(r).decode('ascii')
+
+    def key_by_name(self, name):
+        r = lib.xkb_keymap_key_by_name(self._keymap, name.encode('ascii'))
+        if r == lib.XKB_KEYCODE_INVALID:
+            raise XKBKeyDoesNotExist(name)
+        return r
+
     def num_mods(self):
         """Get the number of modifiers in the keymap."""
         return lib.xkb_keymap_num_mods(self._keymap)
@@ -574,7 +655,7 @@ class Keymap:
         return lib.xkb_keymap_key_repeats(self._keymap, key) == 1
 
     # Keyboard State http://xkbcommon.org/doc/current/group__state.html
-    
+
     def state_new(self):
         """Create a new keyboard state object.
 
@@ -586,17 +667,20 @@ class Keymap:
 
 @enum.unique
 class KeyDirection(enum.IntEnum):
+    "Specifies the direction of the key (press / release)"
     XKB_KEY_UP = lib.XKB_KEY_UP
     XKB_KEY_DOWN = lib.XKB_KEY_DOWN
 
 
-XKB_KEY_UP = KeyDirection.XKB_KEY_UP
-XKB_KEY_DOWN = KeyDirection.XKB_KEY_DOWN
-
-
 @enum.unique
-class StateComponent(enum.IntFlag):
-    "An integer corresponding to enum xkb_state_component"
+class StateComponent(_IntFlag):
+    """Modifier and layout types for state objects
+
+    This enum is bitmaskable, e.g. (XKB_STATE_MODS_DEPRESSED |
+    XKB_STATE_MODS_LATCHED) is valid to exclude locked modifiers.
+
+    In XKB, the DEPRESSED components are also known as 'base'.
+    """
     XKB_STATE_MODS_DEPRESSED = lib.XKB_STATE_MODS_DEPRESSED
     XKB_STATE_MODS_LATCHED = lib.XKB_STATE_MODS_LATCHED
     XKB_STATE_MODS_LOCKED = lib.XKB_STATE_MODS_LOCKED
@@ -607,18 +691,43 @@ class StateComponent(enum.IntFlag):
     XKB_STATE_LAYOUT_EFFECTIVE = lib.XKB_STATE_LAYOUT_EFFECTIVE
     XKB_STATE_LEDS = lib.XKB_STATE_LEDS
 
-for _sc in StateComponent:
-    globals()[_sc.name] = _sc
 
 @enum.unique
-class StateMatch(enum.IntFlag):
-    "An integer corresponding to enum xkb_state_match"
+class StateMatch(_IntFlag):
+    """State match flags
+
+    Match flags for KeyboardState.mod_indices_are_active() and
+    KeyboardState.mod_names_are_active(), specifying the conditions
+    for a successful match.
+
+    XKB_STATE_MATCH_NON_EXCLUSIVE is bitmaskable with the other modes.
+    """
     XKB_STATE_MATCH_ANY = lib.XKB_STATE_MATCH_ANY
     XKB_STATE_MATCH_ALL = lib.XKB_STATE_MATCH_ALL
     XKB_STATE_MATCH_NON_EXCLUSIVE = lib.XKB_STATE_MATCH_NON_EXCLUSIVE
 
-for _sc in StateMatch:
-    globals()[_sc.name] = _sc
+
+@enum.unique
+class ConsumedMode(enum.IntEnum):
+    """Consumed modifiers mode
+
+    There are several possible methods for deciding which modifiers
+    are consumed and which are not, each applicable for different
+    systems or situations. The mode selects the method to use.
+
+    Keep in mind that in all methods, the keymap may decide to
+    "preserve" a modifier, meaning it is not reported as consumed even
+    if it would have otherwise.
+    """
+    XKB_CONSUMED_MODE_XKB = lib.XKB_CONSUMED_MODE_XKB
+    XKB_CONSUMED_MODE_GTK = lib.XKB_CONSUMED_MODE_GTK
+
+
+# Global names for enum members
+for _ec in (KeyDirection, StateComponent, StateMatch, ConsumedMode):
+    for _sc in _ec:
+        globals()[_sc.name] = _sc
+
 
 class KeyboardState:
     def __init__(self, keymap):
@@ -641,7 +750,7 @@ class KeyboardState:
         released.
 
         This entry point is intended for programs which track the
-        keyboard state explictly (like an evdev client). If the state
+        keyboard state explicitly (like an evdev client). If the state
         is serialized to you by a master process (like a Wayland
         compositor) using functions like KeyboardState.serialize_mods(),
         you should use KeyboardState.update_mask() instead. The two
@@ -667,7 +776,7 @@ class KeyboardState:
         """
         return StateComponent(
             lib.xkb_state_update_key(self._state, key, direction))
-                                        
+
     def update_mask(self, depressed_mods, latched_mods, locked_mods,
                     depressed_layout, latched_layout, locked_layout):
         """Update a keyboard state from a set of explicit masks.
@@ -719,7 +828,7 @@ class KeyboardState:
         for i in range(0, r):
             syms.append(syms_out[0][i])
         return syms
-        
+
     def key_get_string(self, key):
         """Get the Unicode/UTF-8 string obtained from pressing a particular
         key in a given keyboard state.
@@ -769,10 +878,10 @@ class KeyboardState:
         and layout.
 
         layout must be smaller than:
-        State.get_keymap().num_layouts_for_key(key) 
+        State.get_keymap().num_layouts_for_key(key)
 
         Usually layout would be:
-        State.key_get_layout(key) 
+        State.key_get_layout(key)
 
         Returns the shift level index. If the key or layout are
         invalid, raises XKBInvalidLayoutIndex.
@@ -886,12 +995,14 @@ class KeyboardState:
         """
         args = [ffi.cast("int", x) for x in mods]
         args.append(ffi.cast("int", lib.XKB_MOD_INVALID))
-        r = lib.xkb_state_mod_indices_are_active(self._state, type, match, *args)
+        r = lib.xkb_state_mod_indices_are_active(
+            self._state, type, match, *args)
         if r == -1:
             raise XKBInvalidModifierIndex()
         return r == 1
 
-    def mod_index_is_consumed(self, key, idx):
+    def mod_index_is_consumed(self, key, idx,
+                              mode=ConsumedMode.XKB_CONSUMED_MODE_XKB):
         """Test whether a modifier is consumed by keyboard state translation
         for a key.
 
@@ -899,7 +1010,7 @@ class KeyboardState:
         If the modifier index is not valid in the keymap, raises
         XKBInvalidModifierIndex.
         """
-        r = lib.xkb_state_mod_index_is_consumed(self._state, key, idx)
+        r = lib.xkb_state_mod_index_is_consumed2(self._state, key, idx, mode)
         if r == -1:
             raise XKBInvalidModifierIndex()
         return r == 1
@@ -912,12 +1023,13 @@ class KeyboardState:
         """
         return lib.xkb_state_mod_mask_remove_consumed(self._state, key, mask)
 
-    def key_get_consumed_mods(self, key):
+    def key_get_consumed_mods(self, key,
+                              mode=ConsumedMode.XKB_CONSUMED_MODE_XKB):
         """Get the mask of modifiers consumed by translating a given key.
 
         Returns a mask of the consumed modifiers.
         """
-        return lib.xkb_state_key_get_consumed_mods(self._state, key)
+        return lib.xkb_state_key_get_consumed_mods2(self._state, key, mode)
 
     def layout_name_is_active(self, name, type):
         """Test whether a layout is active in a given keyboard state by name.
@@ -929,7 +1041,8 @@ class KeyboardState:
         If multiple layouts in the keymap have this name, the one with
         the lowest index is tested.
         """
-        r = lib.xkb_state_layout_name_is_active(self._state, name.encode(), type)
+        r = lib.xkb_state_layout_name_is_active(
+            self._state, name.encode(), type)
         if r == -1:
             raise XKBLayoutDoesNotExist(name)
         return r == 1
